/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <gpdef.h>
#include <gpstdlib.h>
#include <gpgraphic.h>
#include <gpstdio.h>
#include <gpfont.h>
#include <gpmm.h>

#include "m6809.h"
#include "pia.h"
#include "sound.h"
#include "video.h"
#include "types.h"

//#define SAMPLE_RATE 22418  /* PCLK = 33MHz */
#define SAMPLE_RATE  22135  /* PCLK = 50MHz */
#define SAMPLE_SIZE 2
#define FRAME_SIZE 512
#define BUFFER_SIZE (FRAME_SIZE * 2)
#define SAMPLE_CYCLES (OSCILLATOR_RATE / SAMPLE_RATE)
//#define BUFFER_CYCLES ((OSCILLATOR_RATE * BUFFER_SIZE) / SAMPLE_RATE)
#define BUFFER_CYCLES ((OSCILLATOR_RATE / SAMPLE_RATE) * BUFFER_SIZE)
#define FRAME_CYCLES ((OSCILLATOR_RATE / SAMPLE_RATE) * FRAME_SIZE)

extern Uint32 next_sound_update;

static Uint32 sound_cycle_base;
static Uint16 *buffer;
static Uint16 *buffer_frame1;
static Uint16 *buffer_end;
static Uint16 *wrptr;
static Uint32 lastsample;
static volatile int idx_buf;
static volatile Uint32 *playpos;

int sound_init(void) {
	GpPcmInit(PCM_S22, PCM_8BIT);
	buffer = (Uint16 *)malloc(BUFFER_SIZE * SAMPLE_SIZE);
	buffer_frame1 = buffer + FRAME_SIZE;
	buffer_end = buffer + BUFFER_SIZE;
	return 0;
}

void sound_reset(void) {
	GpPcmStop();
	memset(buffer, 0x80, BUFFER_SIZE * SAMPLE_SIZE);
	wrptr = buffer;
	sound_cycle_base = current_cycle;
	next_sound_update = sound_cycle_base + FRAME_CYCLES + 320;
	lastsample = 0x8080;
	GpPcmPlay((Uint16 *)buffer, BUFFER_SIZE * SAMPLE_SIZE, 1);
	GpPcmLock((Uint16 *)buffer, (int *)&idx_buf, (Uint32 **)&playpos);
}

void sound_update(void) {
	Uint32 elapsed_cycles = current_cycle - sound_cycle_base;
	Uint32 fill_with;
	Uint16 *fill_to;
	if (elapsed_cycles > BUFFER_CYCLES * 2) {
		sound_cycle_base = current_cycle;
		elapsed_cycles = 0;
	}
	fill_to = buffer + (elapsed_cycles/(Uint32)SAMPLE_CYCLES);
	if (!(PIA_1B.control_register & 0x08)) {
		/* Single-bit sound */
		//fill_with = (PIA_1B.port_output & 0x02) ? 0xffff : 0;
		fill_with = lastsample;
	} else  {
		if (PIA_0B.control_register & 0x08 || PIA_0A.control_register & 0x08) {
			/* Sound disabled */
			fill_with = lastsample;
		} else {
			/* DAC output */
			fill_with = PIA_1A.port_output & 0xfc;
			fill_with |= (fill_with << 8);
		}
	}
	if (wrptr < buffer_frame1) {
		while (*playpos < (Uint32)buffer_frame1);
		next_sound_update = sound_cycle_base + FRAME_CYCLES + 320;
		while (wrptr < fill_to && wrptr < buffer_frame1)
			*(wrptr++) = lastsample;
		if (wrptr >= buffer_frame1) {
			while (*playpos >= (Uint32)buffer_frame1);
			next_sound_update = sound_cycle_base + BUFFER_CYCLES + 320;
			while (wrptr < fill_to)
				*(wrptr++) = lastsample;
		}
	} else {
		while (*playpos >= (Uint32)buffer_frame1);
		next_sound_update = sound_cycle_base + BUFFER_CYCLES + 320;
		while (wrptr < fill_to && wrptr < buffer_end)
			*(wrptr++) = lastsample;
		if (wrptr >= buffer_end) {
			while (*playpos < (Uint32)buffer_frame1);
			fill_to -= BUFFER_SIZE;
			sound_cycle_base += BUFFER_CYCLES;
			next_sound_update = sound_cycle_base + FRAME_CYCLES + 320;
			wrptr = buffer;
			while (wrptr < fill_to)
				*(wrptr++) = lastsample;
		}
	}
	lastsample = fill_with;
	return;
}
